
		/*********************************************************
		** Decides if the optic nerve (ON) is present in the image
		** or not, and if it is, locates it.
		**
		** This version takes the vessels.net map and finds its
		** fuzzy convergence.  The net is broken up into
		** segments, where each segment contains two
		** endpoints only (but is not necessarily a straight line).
		** Breaks are made at all branch points.
		** Short segments are discarded.
		** Segments are modeled by 2nd-order parametric curves,
		** instead of straight lines.
		**
		** Coded by Adam Hoover in November '97.
		**
		** Ported to Windows in Sept '00.
		*********************************************************/

#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "globals.h"




void LocateOpticNerve(unsigned char *EqualizedImage,			/* input can be NULL, meaning don't use */
					  int EqBPP,								/* equalized image can be 3 BPP (RGB) or 1 BPP (greyscale) */
					  unsigned char *VesselsImage,				/* input can be NULL, meaning don't use */
					  int ROWS, int COLS,						/* size of images */
					  char *VesselsPath,						/* if not NULL, read vessels images from this path */
					  char *ImagePrefix,						/* with path, read ImagePrefix-vesselsN for N=0...5 */
					  unsigned char *NerveImage)				/* output, for display */

{
char			filename[320];
unsigned char	*SearchImage,*ThreshImage,*VesselsNImage;
int				cumhist[256],a,i,j,k,r,c,r2,c2,n;
int				*queue,*indices;
int				RegionSizes[1000];
int				RegionLabels[1000];
int				TotalRegions;
int				largest,temp;
double			set1avg,set2avg,set1var,set2var,fisher[1000];
int				expand,total_pixels;
int				PointRow[7],PointCol[7];


if ((SearchImage=(unsigned char *)calloc(ROWS*COLS,1)) == NULL  ||
	(ThreshImage=(unsigned char *)calloc(ROWS*COLS,1)) == NULL  ||
	(indices=(int *)calloc(ROWS*COLS,sizeof(int))) == NULL  ||
	(queue=(int *)calloc(ROWS*COLS,sizeof(int))) == NULL)
  {
  MessageBox(NULL,"Unable to allocate memory.","LocateOpticNerve()",MB_OK | MB_APPLMODAL);
  return;
  }

for (n=0; n<7; n++)
  {
  if (n < 6  &&  VesselsPath != NULL)
	{
	sprintf(filename,"%s\\%s-vessels%d.ppm",VesselsPath,ImagePrefix,n);
	if (ReadImage(filename,&VesselsNImage,&r,&c) != 1)
	  {
	  PointRow[n]=PointCol[n]=-1;	  /* location found for image n is NULL */
	  continue;
	  }
		/* Run vessels image through fuzzy convergence */
    FuzzyConvergence(VesselsNImage,ROWS,COLS,SearchImage);
	free(VesselsNImage);	/* was allocated inside ReadImage() */
	}
  else if (n == 0  &&  VesselsPath == NULL  &&  VesselsImage != NULL)
	{	/* no path, so look for single vessels image supplied at function call */
    FuzzyConvergence(VesselsImage,ROWS,COLS,SearchImage);
	}
  else if (n == 6  &&  EqualizedImage != NULL)
	{	/* Search for bright spot in equalized (raw) image */
	for (i=0; i<ROWS*COLS; i++)
	  SearchImage[i]=EqualizedImage[i*EqBPP+(EqBPP/3)];
	}
  else
	{	/* no point available for this value of n */
	PointRow[n]=PointCol[n]=-1;
	continue;
	}

	/* The following operations find the brightest spot for either convergence or equalized image */

		/* Threshold image for brightest 1500 pixels */
  for (i=0; i<256; i++)
    cumhist[i]=0;
  for (i=0; i<ROWS*COLS; i++)
    cumhist[SearchImage[i]]++;
  for (i=254; i>=0; i--)
    cumhist[i]+=cumhist[i+1];
  for (i=255; i>=0; i--)
    if (cumhist[i] >= 1500)
	  break;
  if (i <= 0)		  /* empty image */
	{
	PointRow[n]=PointCol[n]=-1;
	continue;
	}
  for (j=0; j<ROWS*COLS; j++)
    if (SearchImage[j] >= i)
	  {
	  ThreshImage[j]=SearchImage[j]=255;
	  }
    else
	  {
	  ThreshImage[j]=SearchImage[j]=0;
	  }

		/* Expand regions 5 pixels outward to connect across vessel-caused gaps */
  for (expand=0; expand<5; expand++)
	{
    total_pixels=0;
    for (r=1; r<ROWS-1; r++)
	  for (c=1; c<COLS-1; c++)
	    if (SearchImage[r*COLS+c] == 0)
		  {
		  if (SearchImage[(r-1)*COLS+c]  ||  SearchImage[(r+1)*COLS+c]  ||  SearchImage[r*COLS+c-1]  ||  SearchImage[r*COLS+c+1])
			{
		    indices[total_pixels]=r*COLS+c;
		    total_pixels++;
			} 
		  }
    for (i=0; i<total_pixels; i++)
	  SearchImage[indices[i]]=255;
	}

		/* Find sizes of the connected components */
  TotalRegions=0;
  for (i=0; i<ROWS*COLS; i++)
    if (SearchImage[i] == 255)
	  {
	  j=PaintFill(SearchImage,ROWS,COLS,i/COLS,i%COLS,255,254-TotalRegions,queue,ROWS*COLS,indices);
	  RegionSizes[TotalRegions]=0;	/* painted image is just for connecting -- real size is in ThreshImage */
	  for (r=0; r<ROWS*COLS; r++)
	    if (SearchImage[r] == 254-TotalRegions  &&  ThreshImage[r] == 255)
		  RegionSizes[TotalRegions]++;
	  RegionLabels[TotalRegions]=254-TotalRegions;
	  TotalRegions++;
	  if (TotalRegions >= 250)
		{
	    MessageBox(NULL,"TotalRegions exceeded!","LocateOpticNerve()",MB_OK | MB_APPLMODAL);
	    return;
		}
	  }

		/* Sort the sizes */
  for (i=0; i<TotalRegions; i++)
	{
    largest=i;
    for (j=i+1; j<TotalRegions; j++)
	  if (RegionSizes[j] > RegionSizes[largest])
		{
	    temp=RegionSizes[i];
	    RegionSizes[i]=RegionSizes[j];
	    RegionSizes[j]=temp;
	    temp=RegionLabels[i];
	    RegionLabels[i]=RegionLabels[j];
	    RegionLabels[j]=temp;
		}
	}

		/* Calculate Fisher's linear discriminant of sizes */
  for (a=0; a<TotalRegions-1; a++)
	{
    set1avg=set1var=set2avg=set2var=0.0;
    for (i=0; i<=a; i++)
	  set1avg+=(double)RegionSizes[i];
    set1avg/=i;
    for (; i<TotalRegions; i++)
	  set2avg+=(double)RegionSizes[i];
    set2avg/=(TotalRegions-a-1);
    for (i=0; i<=a; i++)
	  set1var+=(double)SQR(set1avg-RegionSizes[i]);
    set1var/=i;
    for (; i<TotalRegions; i++)
	  set2var+=(double)SQR(set2avg-RegionSizes[i]);
    set2var/=(TotalRegions-a-1);
    fisher[a]=SQR(set1avg-set2avg)/(set1var+set2var);
    if (a == 0)
	  largest=0;
    else if (fisher[a] > fisher[largest])
	  largest=a;
	}


  if (largest == 0)
	{	/* Copy largest region into output image, and identify its position */
    r=c=a=0;
    for (i=0; i<ROWS*COLS; i++)
      if (SearchImage[i] == RegionLabels[0])
		{
	    NerveImage[i]=255;
	    r+=(i/COLS); c+=(i%COLS);
	    a++;
		}
    PointRow[n]=r/a; PointCol[n]=c/a;	  /* center of spot thought to be nerve */
	}
  else
	{
	PointRow[n]=PointCol[n]=-1;			  /* multiple spots = don't know */
	}

  }

		/* See if any 3 of the values from the 6 fuzzy convergence images agree on point */
for (i=0; i<4; i++)
  {
  if (PointRow[i] == -1)
	continue;
  for (j=i+1; j<5; j++)
	{
	if (PointRow[j] == -1)
	  continue;
	for (k=j+1; k<6; k++)
	  {
	  if (PointRow[k] == -1)
		continue;
	  if (sqrt(SQR(PointRow[i]-PointRow[j])+SQR(PointCol[i]-PointCol[j])) < (double)NerveRadius  &&
		  sqrt(SQR(PointRow[j]-PointRow[k])+SQR(PointCol[j]-PointCol[k])) < (double)NerveRadius  &&
		  sqrt(SQR(PointRow[i]-PointRow[k])+SQR(PointCol[i]-PointCol[k])) < (double)NerveRadius)
		break;
	  }
	if (k < 6)
	  break;
	}
  if (j < 5)
	break;
  }
if (i < 4)
  {		/* found agreeing set in convergence images (i,j,k) */
  NerveRow=(PointRow[i]+PointRow[j]+PointRow[k])/3;
  NerveCol=(PointCol[i]+PointCol[j]+PointCol[k])/3;
  }
else if (PointRow[6] != -1)
  {		/* use bright point as fall-back */
  NerveRow=PointRow[6];
  NerveCol=PointCol[6];
  }
else
  NerveRow=NerveCol=-1;

if (0 /* NerveRow != -1 */)
  {
  if (EqualizedImage != NULL)
	{	/* copy equalized image into result image */
    for (r=0; r<ROWS; r++)
	  for (c=0; c<COLS; c++)
	    if (EqBPP == 1)
		  NerveImage[r*COLS+c]=EqualizedImage[r*COLS+c];
	    else
		  NerveImage[r*COLS+c]=(unsigned char)((int)(EqualizedImage[(r*COLS+c)*3+0]+EqualizedImage[(r*COLS+c)*3+1]+
													  EqualizedImage[(r*COLS+c)*3+2])/3);
	}

		/* Mark location of optic nerve with 'X' */
  for (r=NerveRow-5,r2=-5; r<=NerveRow+5; r++,r2++)
    for (c=NerveCol-5,c2=-5; c<=NerveCol+5; c++,c2++)
	  {
      if (abs(abs(r2)-abs(c2)) > 3  ||  r < 0  ||  r >= ROWS  ||  c < 0  ||  c >= COLS)
        continue;
      NerveImage[r*COLS+c]=180;
	  }
  }

free(SearchImage);
free(ThreshImage);
free(indices);
free(queue);
}







#define MAX_QUEUE	500000	  /* constant for allocating space for arrays */
#define MAX_POINTS	10000	  /* constant for allocating space for arrays */
#define MAX_RADIUS	60		  /* constant for allocating space for arrays */
#define MAX_SEGS	1000	  /* constant for allocating space for arrays */

void FuzzyConvergence(unsigned char *VesselsImage,
				 	   int ROWS, int COLS,
					   unsigned char *FuzzyImage)		  /* output */
					/* also sets NerveRow and NerveCol global vars */
{
unsigned char	*NetImage,*VoteImage;
int				*ConvergenceImage,*MeltImage;
int				*Queue;	  /* used in PaintFill(); allocated here, only once */
int				a,i,j,r,c,r2,c2;
int				*indices,count,MinVesselSize,FuzzyRadius,OnlySegment,Extension;
double			FuzzyFactor;
int				FILTER_SIZE;
int				endpoints[MAX_POINTS],total_ends;
int				branches[MAX_POINTS],total_branches;
int				Size,ei[MAX_SEGS][5];
int				*circle1_indices,*circle2_indices,*circle3_indices;
int				circle1_count,circle2_count,circle3_count;
int				*curve_indices,curve_count,p1x,p1y,p2x,p2y,p3x,p3y;
int				x1,y1,x2,y2,x3,y3,s;
double			angle;
int				biggest,biggest_index,conv;
int				TotalSegments;


MinVesselSize=0; // 15;	  /* delete segments with less pixels than this count */
FuzzyRadius=40;		  /* "R" in fuzzy segment equation (radius of set of lines) */
FuzzyFactor=2.5;	  /* scaling factor used if FuzzyRadius is adapted to segment size */
Extension=15;		  /* distance, in pixels, segment is extended at each end before fuzzification */
OnlySegment=-1;		  /* for debugging; set to ID to see fuzzy segment of only that label */
FILTER_SIZE=11;		  /* N x N window for smoothing (mean filter) */

		/*********************************************************
		** Allocate memory for images.
		*********************************************************/

if ((NetImage=(unsigned char *)calloc(1,ROWS*COLS*1)) == NULL  ||
	(VoteImage=(unsigned char *)calloc(1,ROWS*COLS*1)) == NULL  ||
	(MeltImage=(int *)calloc(sizeof(int),ROWS*COLS*1)) == NULL  ||
	(Queue=(int *)calloc(sizeof(int),MAX_QUEUE)) == NULL  ||
	(ConvergenceImage=(int *)calloc(sizeof(int),ROWS*COLS)) == NULL)
  {
  MessageBox(NULL,"FindOpticNerve();","Unable to allocate memory for images.\n",MB_OK | MB_APPLMODAL);
  exit(0);
  }
count=0;
for (i=0; i<ROWS*COLS; i++)
  {
  NetImage[i]=VesselsImage[i];	  /* copy original */
  if (NetImage[i])
	count++;
  }
indices=(int *)calloc(count,sizeof(int));
count=0;
for (i=0; i<ROWS*COLS; i++)
  if (NetImage[i])
	indices[count++]=i;

        /*********************************************************
        ** Thin vessels image down to minimum 8-connectedness.
		** Count up indices of vessels points after thinning.
		** Find branchpoints and endpoints of thinned net.
		*********************************************************/

ThinPixels(NetImage,ROWS,COLS,indices,count,255);
free(indices);
count=0;
for (i=0; i<ROWS*COLS; i++)
  if (NetImage[i])
    count++;
indices=(int *)calloc(count,sizeof(int));
count=0;
for (i=0; i<ROWS*COLS; i++)
  if (NetImage[i])
    {
    indices[count]=i;
    count++;
    }
BranchesAndEnds(NetImage,ROWS,COLS,indices,count,endpoints,&total_ends,branches,&total_branches,MAX_POINTS,255);

        /*********************************************************
        ** Unlabel all branch points to break up everything.
        *********************************************************/

for (i=0; i<total_branches; i++)
  {
  for (r2=branches[i]/COLS-1; r2<=branches[i]/COLS+1; r2++)
    for (c2=branches[i]%COLS-1; c2<=branches[i]%COLS+1; c2++)
      NetImage[r2*COLS+c2]=0;
  }

        /*********************************************************
        ** Go through segments one at a time, removing small ones
		** and storing endpoints of large ones.
		*********************************************************/

TotalSegments=0;
for (r=0; r<ROWS; r++)
  for (c=0; c<COLS; c++)
    {
    if (NetImage[r*COLS+c] != 255)
      continue;
    Size=PaintFill(NetImage,ROWS,COLS,r,c,255,254,Queue,MAX_QUEUE,indices);
	if (Size < MinVesselSize)
	  {
	  for (i=0; i<Size; i++)
		NetImage[indices[i]]=0;
	  continue;
	  }
    BranchesAndEnds(NetImage,ROWS,COLS,indices,Size,endpoints,&total_ends,branches,&total_branches,MAX_POINTS,254);
	if (total_ends != 2)
	  {	/* error of some kind -- skip this segment */
	  for (i=0; i<Size; i++)
		NetImage[indices[i]]=0;
	  continue;
	  }
	for (i=0; i<total_ends; i++)
	  ei[TotalSegments][i]=endpoints[i];
	TotalSegments++;
	if (TotalSegments >= MAX_SEGS)
	  {
	  MessageBox(NULL,"MAX_SEGS exceeded","FindOpticNerve();",MB_OK | MB_APPLMODAL);
	  exit(0);
	  }
	for (i=0; i<Size; i++)
	  NetImage[indices[i]]=253;
	}
free(indices);

        /*********************************************************
		** Find all mid-points of segments.
		*********************************************************/

curve_indices=(int *)calloc(MAX_POINTS,sizeof(int));
for (i=0; i<TotalSegments; i++)
  {
  curve_count=TraceSegment(NetImage,ROWS,COLS,ei[i][0]/COLS,ei[i][0]%COLS,Queue,MAX_QUEUE,curve_indices);
  ei[i][2]=curve_indices[curve_count/2];
  }
free(curve_indices);


        /*********************************************************
        ** Create a fuzzy intersection image.
		*********************************************************/

circle1_indices=(int *)calloc(MAX_RADIUS*8,sizeof(int));
circle2_indices=(int *)calloc(MAX_RADIUS*8,sizeof(int));
circle3_indices=(int *)calloc(MAX_RADIUS*8,sizeof(int));
curve_indices=(int *)calloc(2000,sizeof(int));
for (i=0; i<TotalSegments; i++)
  {
  if (OnlySegment != -1  &&  i != OnlySegment)
    continue;
  p1x=ei[i][0]%COLS; p1y=ei[i][0]/COLS;
  p2x=ei[i][2]%COLS; p2y=ei[i][2]/COLS;
  p3x=ei[i][1]%COLS; p3y=ei[i][1]/COLS;
  MakeCurveIndices(p1x,p1y,p2x,p2y,p3x,p3y,ROWS,COLS,Extension,curve_indices,&curve_count);
  p1x=curve_indices[0]%COLS; p1y=curve_indices[0]/COLS;
  p2x=curve_indices[curve_count/2]%COLS; p2y=curve_indices[curve_count/2]/COLS;
  p3x=curve_indices[curve_count-1]%COLS; p3y=curve_indices[curve_count-1]/COLS;
/*
  FuzzyRadius=(int)(sqrt((double)(SQR(p3x-p1x)+SQR(p3y-p1y)))/FuzzyFactor);
  if (FuzzyRadius > MAX_RADIUS) FuzzyRadius=MAX_RADIUS;
  if (FuzzyRadius < 5) FuzzyRadius=5;
*/
  MakeCircleIndices((double)p1x,(double)p1y,(double)FuzzyRadius,ROWS,COLS,circle1_indices,&circle1_count);
  MakeCircleIndices((double)p2x,(double)p2y,(double)FuzzyRadius,ROWS,COLS,circle2_indices,&circle2_count);
  MakeCircleIndices((double)p3x,(double)p3y,(double)FuzzyRadius,ROWS,COLS,circle3_indices,&circle3_count);
/*
if (circle1_count != circle2_count  ||  circle2_count != circle3_count)
  { printf("unequal counts: %d %d %d\n",circle1_count,circle2_count,
	circle3_count); exit(0); }
*/
  angle=atan2((double)(p3y-p1y),(double)(p3x-p1x));
  s=(int)(((M_PI/2.0+angle)/(2.0*M_PI))*(double)circle1_count);
  for (a=0; a<circle1_count; a++)	/* a => first point of arc */
    {
    y1=circle1_indices[(s+circle1_count+a)%circle1_count]/COLS;
    x1=circle1_indices[(s+circle1_count+a)%circle1_count]%COLS;
    y2=circle2_indices[(s+circle2_count*3/2-a)%circle2_count]/COLS;
    x2=circle2_indices[(s+circle2_count*3/2-a)%circle2_count]%COLS;
    y3=circle3_indices[(s+circle1_count+a)%circle3_count]/COLS;
    x3=circle3_indices[(s+circle1_count+a)%circle3_count]%COLS;
    MakeCurveIndices(x1,y1,x2,y2,x3,y3,ROWS,COLS,0,curve_indices,&curve_count);
    for (j=0; j<curve_count; j++)
      {
      if (VoteImage[curve_indices[j]])
        continue;	/* already voted */
      VoteImage[curve_indices[j]]=1;
      ConvergenceImage[curve_indices[j]]++;
/*
      ConvergenceImage[curve_indices[j]]+=(int)(((double)abs(j-curve_count/2)/
		(double)curve_count+0.5)*(double)100.0);
      ConvergenceImage[curve_indices[j]]+=(int)(((double)abs(j-curve_count/2)/
		(double)curve_count+0.5)*(double)Sizes[i]);
*/
      }
    }
  if (p3y < p1y)
    { j=p3y; p3y=p1y; p1y=j; }
  if (p3x < p1x)
    { j=p3x; p3x=p1x; p1x=j; }
  for (r=p1y-(FuzzyRadius+1); r<=p3y+(FuzzyRadius+1); r++)
    for (c=p1x-(FuzzyRadius+1); c<=p3x+(FuzzyRadius+1); c++)
      if (r >= 0  &&  r < ROWS  &&  c >= 0  &&  c < COLS)
        VoteImage[r*COLS+c]=0;
  }
free(circle1_indices);
free(circle2_indices);
free(circle3_indices);
free(curve_indices);

		/*********************************************************
		** Smooth convergence image, and find pixel with highest vote.
		** The smoothing is done using a separable filter, using
		** MeltImage as a temporary holding place, and ending with
		** the result back in ConvergenceImage.
		*********************************************************/

biggest=0;
for (r=FILTER_SIZE/2; r<ROWS-FILTER_SIZE/2; r++)
  {
  conv=0;
  for (c=0; c<FILTER_SIZE; c++)
    conv+=ConvergenceImage[r*COLS+c];
  for (c=FILTER_SIZE/2; c<COLS-1-FILTER_SIZE/2; c++)
    {
    MeltImage[r*COLS+c]=conv;
    conv=conv-ConvergenceImage[r*COLS+c-FILTER_SIZE/2]+ConvergenceImage[r*COLS+c+FILTER_SIZE/2+1];
    }
  }
for (c=FILTER_SIZE/2; c<COLS-FILTER_SIZE/2; c++)
  {
  conv=0;
  for (r=0; r<FILTER_SIZE; r++)
    conv+=MeltImage[r*COLS+c];
  for (r=FILTER_SIZE/2; r<ROWS-1-FILTER_SIZE/2; r++)
    {
    ConvergenceImage[r*COLS+c]=conv;
    if (conv > biggest)
      {
      biggest=conv;
      biggest_index=r*COLS+c;
      }
    conv=conv-MeltImage[(r-FILTER_SIZE/2)*COLS+c]+MeltImage[(r+1+FILTER_SIZE/2)*COLS+c];
    }
  }

		/*********************************************************
		** Scale fuzzy image for display.
		*********************************************************/

if (biggest == 0) biggest=1;
for (i=0; i<ROWS*COLS; i++)
  {
		  /* uncomment these lines to include thinned vessels in image, for figures */
/*  if (!(NetImage[i]))
    FuzzyImage[i]=(unsigned char)((double)ConvergenceImage[i]/(double)biggest*255.0);
  else
	FuzzyImage[i]=NetImage[i];
/**/
    FuzzyImage[i]=(unsigned char)((double)ConvergenceImage[i]/(double)biggest*255.0);
/**/
  }
NerveRow=biggest_index&COLS;
NerveCol=biggest_index/COLS;

		/*********************************************************
		** Free up memory that was dynamically allocated.
		*********************************************************/

free(NetImage);
free(VoteImage);
free(ConvergenceImage);
free(MeltImage);
free(Queue);
}






		/*
		** Equalize the illumination in the retinal image.
		** The method calculates the local intensity average using all pixels,
		** and uses average grey (128) as the global intensity average.
		** Every pixel is adjusted by its local average to force it towards
		** the global average.
		*/

void EqualizeIllumination(unsigned char *RawImage,		  /* original retinal image */
						  int ROWS, int COLS,
						  int BPP,						  /* bytes per pixel -- RGB = 3 or greyscale = 1 */
						  unsigned char *AreaImage)		  /* output */

{
int			  i,r,c,r2,c2,total,total_pixels,global_average,radius;
unsigned char *MaskImage,*LocalImage,*AvgImage;
char		  text[80];
HDC			  hDC;
int			  MinPixelCount,MaxPixelCount;


MinPixelCount=30;
MaxPixelCount=50;

MaskImage=(unsigned char *)calloc(ROWS*COLS,1);
LocalImage=(unsigned char *)calloc(ROWS*COLS,1);		  /* raw local intensity average */
AvgImage=(unsigned char *)calloc(ROWS*COLS,1);			  /* smoothed local intensity average */
MakeStareMask(MaskTopRow,MaskBottomRow,MaskCircle,ROWS,COLS,MaskImage);

		/* calculate global intensity */
global_average=128;

		/* calculate local intensity for each pixel */
for (r=0; r<ROWS; r++)
  {
  sprintf(text,"Equalizing...  %d%% complete",(int)(100.0*(double)r/(double)ROWS));
  hDC=GetDC(MainWnd);
  TextOut(hDC,0,0,text,strlen(text));
  ReleaseDC(MainWnd,hDC);
  radius=0;	  /* must start over at each row */
  if (MaskImage[r*COLS+0] != 0)	/* create starting window (1x1 in size) */
	{
	total=RawImage[(r*COLS+0)*BPP+BPP/3];
	total_pixels=1;
	}
  else
	total=total_pixels=0;
  for (c=0; c<COLS; c++)
	{
	if (c > 0)	  /* translate window by one column */
	  {
	  if (c-(radius+1) >= 0)  /* subtract old leftmost column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c-(radius+1)] != 0)
			{
			total-=RawImage[((r+r2)*COLS+c-(radius+1))*BPP+BPP/3];
			total_pixels--;
			}
		  }
		}
	  if (c+radius < COLS)	  /* add in new rightmost column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c+radius]  != 0)
			{
			total+=RawImage[((r+r2)*COLS+c+radius)*BPP+BPP/3];
			total_pixels++;
			}
		  }
		}
	  }

	if (MaskImage[r*COLS+c] == 0)
	  continue;				  /* no local average needed for background pixels */

	while (total_pixels > MaxPixelCount) /* contract window if it contains too many pixels */
	  {
	  if (c-radius >= 0)			/* subtract out left column */
		{
	    for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c-radius] != 0)
			{
		    total-=RawImage[((r+r2)*COLS+c-radius)*BPP+BPP/3];
		    total_pixels--;
			}
		  }
		}
	  if (c+radius < COLS)			/* subtract out right column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c+radius] != 0)
			{
		    total-=RawImage[((r+r2)*COLS+c+radius)*BPP+BPP/3];
		    total_pixels--;
			}
		  }
		}
	  if (r-radius >= 0)			/* subtract out top row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r-radius)*COLS+c+c2] != 0)
			{
			total-=RawImage[((r-radius)*COLS+c+c2)*BPP+BPP/3];
			total_pixels--;
			}
		  }
		}
	  if (r+radius < ROWS)			/* subtract out bottom row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r+radius)*COLS+c+c2] != 0)
			{
			total-=RawImage[((r+radius)*COLS+c+c2)*BPP+BPP/3];
			total_pixels--;
			}
		  }
		}
	  radius--;
	  }		  /* end while() loop to contract window */

	while (total_pixels < MinPixelCount) /* expand window if it contains too few pixels */
	  {
	  radius++;
	  if (radius > 300)			/* failsafe, in case there are less than 1000 masked pixels! */
		break;
	  if (c-radius >= 0)			/* add in new left column */
		{
	    for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c-radius] != 0)
			{
		    total+=RawImage[((r+r2)*COLS+c-radius)*BPP+BPP/3];
		    total_pixels++;
			}
		  }
		}
	  if (c+radius < COLS)			/* add in new right column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c+radius] != 0)
			{
		    total+=RawImage[((r+r2)*COLS+c+radius)*BPP+BPP/3];
		    total_pixels++;
			}
		  }
		}
	  if (r-radius >= 0)			/* add in new top row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r-radius)*COLS+c+c2] != 0)
			{
			total+=RawImage[((r-radius)*COLS+c+c2)*BPP+BPP/3];
			total_pixels++;
			}
		  }
		}
	  if (r+radius < ROWS)			/* add in new bottom row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r+radius)*COLS+c+c2] != 0)
			{
			total+=RawImage[((r+radius)*COLS+c+c2)*BPP+BPP/3];
			total_pixels++;
			}
		  }
		}
	  }		  /* end while() loop to expand window */

	if (total/total_pixels > 255)
      LocalImage[r*COLS+c]=255;
    else
      LocalImage[r*COLS+c]=total/total_pixels;
	}
  }


for (r=0; r<ROWS*COLS; r++)
  AvgImage[r]=LocalImage[r];
if (1) {
		/* smooth the local intensity image */
for (r=0; r<ROWS; r++)
  {
  sprintf(text,"Smoothing...  %d%% complete",(int)(100.0*(double)r/(double)ROWS));
  hDC=GetDC(MainWnd);
  TextOut(hDC,0,0,text,strlen(text));
  ReleaseDC(MainWnd,hDC);
  radius=0;	  /* must start over at each row */
  if (MaskImage[r*COLS+0])	/* create starting window (1x1 in size) */
	{
	total=LocalImage[r*COLS+0];
	total_pixels=1;
	}
  else
	total=total_pixels=0;
  for (c=0; c<COLS; c++)
	{
	if (c > 0)	  /* translate window by one column */
	  {
	  if (c-(radius+1) >= 0)  /* subtract old leftmost column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c-(radius+1)])
			{
			total-=LocalImage[(r+r2)*COLS+c-(radius+1)];
			total_pixels--;
			}
		  }
		}
	  if (c+radius < COLS)	  /* add in new rightmost column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c+radius])
			{
			total+=LocalImage[(r+r2)*COLS+c+radius];
			total_pixels++;
			}
		  }
		}
	  }

	if (MaskImage[r*COLS+c] == 0)
	  continue;				  /* no local average needed for background pixels */

	while (radius > MaxPixelCount) /* contract window if it contains too many pixels */
	  {
	  if (c-radius >= 0)			/* subtract out left column */
		{
	    for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c-radius])
			{
		    total-=LocalImage[(r+r2)*COLS+c-radius];
		    total_pixels--;
			}
		  }
		}
	  if (c+radius < COLS)			/* subtract out right column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c+radius])
			{
		    total-=LocalImage[(r+r2)*COLS+c+radius];
		    total_pixels--;
			}
		  }
		}
	  if (r-radius >= 0)			/* subtract out top row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r-radius)*COLS+c+c2])
			{
			total-=LocalImage[(r-radius)*COLS+c+c2];
			total_pixels--;
			}
		  }
		}
	  if (r+radius < ROWS)			/* subtract out bottom row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r+radius)*COLS+c+c2])
			{
			total-=LocalImage[(r+radius)*COLS+c+c2];
			total_pixels--;
			}
		  }
		}
	  radius--;
	  }		  /* end while() loop to contract window */

	while (radius < MinPixelCount) /* expand window if it contains too few pixels */
	  {
	  radius++;
	  if (c-radius >= 0)			/* add in new left column */
		{
	    for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c-radius])
			{
		    total+=LocalImage[(r+r2)*COLS+c-radius];
		    total_pixels++;
			}
		  }
		}
	  if (c+radius < COLS)			/* add in new right column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c+radius])
			{
		    total+=LocalImage[(r+r2)*COLS+c+radius];
		    total_pixels++;
			}
		  }
		}
	  if (r-radius >= 0)			/* add in new top row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r-radius)*COLS+c+c2])
			{
			total+=LocalImage[(r-radius)*COLS+c+c2];
			total_pixels++;
			}
		  }
		}
	  if (r+radius < ROWS)			/* add in new bottom row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r+radius)*COLS+c+c2])
			{
			total+=LocalImage[(r+radius)*COLS+c+c2];
			total_pixels++;
			}
		  }
		}
	  }		  /* end while() loop to expand window */

	if (total/total_pixels > 255)
      AvgImage[r*COLS+c]=255;
    else
      AvgImage[r*COLS+c]=total/total_pixels;
	}
  }

}


		/* adjust intensity by smoothed local average */
for (i=0; i<BPP; i++)
  for (r=0; r<ROWS; r++)
	{
    for (c=0; c<COLS; c++)
	  {
	  if (!MaskImage[r*COLS+c])
	    continue;
	  if ((int)RawImage[(r*COLS+c)*BPP+i]+global_average-(int)AvgImage[r*COLS+c] > 255)
	    AreaImage[(r*COLS+c)*BPP+i]=255;
	  else if ((int)RawImage[(r*COLS+c)*BPP+i]+global_average-(int)AvgImage[r*COLS+c] < 0)
	    AreaImage[(r*COLS+c)*BPP+i]=0;
	  else
	    AreaImage[(r*COLS+c)*BPP+i]=RawImage[(r*COLS+c)*BPP+i]+global_average-AvgImage[r*COLS+c];
	  }
	}


free(MaskImage);
free(LocalImage);
free(AvgImage);
}








		/*
		** Equalize the illumination in the retinal image.
		** The method uses the blood vessel pixels to estimate a
		** local inensity average, as well as a global intensity average.
		** Every pixel is adjusted by its local average to force it to
		** the global average.
		*/

void EqualizeIlluminationOnVessels(unsigned char *RawImage,		  /* original retinal image */
								   unsigned char *VesselsImage,	  /* reference image */
								   int ROWS, int COLS,
								   unsigned char *AreaImage)		  /* output */

{
int			  i,r,c,r2,c2,total,total_pixels,global_average,radius;
unsigned char *MaskImage,*LocalImage,*AvgImage,*RGBEqualizedImage;
char		  text[80];
HDC			  hDC;
int			  RETURN_LOCAL_AVERAGE_IMAGE=0;
int			  SAVE_RGB_EQUALIZED_IMAGE=0;
int			  MaxRadius=0;
int			  *RadiusImage;


MaskImage=(unsigned char *)calloc(ROWS*COLS,1);
LocalImage=(unsigned char *)calloc(ROWS*COLS,1);
AvgImage=(unsigned char *)calloc(ROWS*COLS,1);
RadiusImage=(int *)calloc(ROWS*COLS,sizeof(int));
MakeStareMask(MaskTopRow,MaskBottomRow,MaskCircle,ROWS,COLS,MaskImage);

		/* calculate global blood vessel intensity */
total=total_pixels=0;
for (r=0; r<ROWS*COLS; r++)
  if (VesselsImage[r])
	{
	total+=RawImage[r*3+1];
	total_pixels++;
	}
global_average=total/total_pixels;

		/* calculate local blood vessel intensity for each pixel */
for (r=0; r<ROWS; r++)
  {
  sprintf(text,"Equalizing...  %d%% complete",(int)(100.0*(double)r/(double)ROWS));
  hDC=GetDC(MainWnd);
  TextOut(hDC,0,0,text,strlen(text));
  ReleaseDC(MainWnd,hDC);
  radius=0;	  /* must start over at each row */
  if (VesselsImage[r*COLS+0])	/* create starting window (1x1 in size) */
	{
	total=RawImage[(r*COLS+0)*3+1];
	total_pixels=1;
	}
  else
	total=total_pixels=0;
  for (c=0; c<COLS; c++)
	{
	if (c > 0)	  /* translate window by one column */
	  {
	  if (c-(radius+1) >= 0)  /* subtract old leftmost column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (VesselsImage[(r+r2)*COLS+c-(radius+1)])
			{
			total-=RawImage[((r+r2)*COLS+c-(radius+1))*3+1];
			total_pixels--;
			}
		  }
		}
	  if (c+radius < COLS)	  /* add in new rightmost column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (VesselsImage[(r+r2)*COLS+c+radius])
			{
			total+=RawImage[((r+r2)*COLS+c+radius)*3+1];
			total_pixels++;
			}
		  }
		}
	  }

	if (MaskImage[r*COLS+c] == 0)
	  continue;				  /* no local average needed for background pixels */

	while (total_pixels > 50) /* contract window if it contains too many pixels */
	  {
	  if (c-radius >= 0)			/* subtract out left column */
		{
	    for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (VesselsImage[(r+r2)*COLS+c-radius])
			{
		    total-=RawImage[((r+r2)*COLS+c-radius)*3+1];
		    total_pixels--;
			}
		  }
		}
	  if (c+radius < COLS)			/* subtract out right column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (VesselsImage[(r+r2)*COLS+c+radius])
			{
		    total-=RawImage[((r+r2)*COLS+c+radius)*3+1];
		    total_pixels--;
			}
		  }
		}
	  if (r-radius >= 0)			/* subtract out top row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (VesselsImage[(r-radius)*COLS+c+c2])
			{
			total-=RawImage[((r-radius)*COLS+c+c2)*3+1];
			total_pixels--;
			}
		  }
		}
	  if (r+radius < ROWS)			/* subtract out bottom row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (VesselsImage[(r+radius)*COLS+c+c2])
			{
			total-=RawImage[((r+radius)*COLS+c+c2)*3+1];
			total_pixels--;
			}
		  }
		}
	  radius--;
	  }		  /* end while() loop to contract window */

	while (total_pixels < 30) /* expand window if it contains too few pixels */
	  {
	  radius++;
	  if (radius > MaxRadius) MaxRadius=radius;
	  if (radius > 300)			/* failsafe, in case there are less than 1000 vessels pixels! */
		break;
	  if (c-radius >= 0)			/* add in new left column */
		{
	    for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (VesselsImage[(r+r2)*COLS+c-radius])
			{
		    total+=RawImage[((r+r2)*COLS+c-radius)*3+1];
		    total_pixels++;
			}
		  }
		}
	  if (c+radius < COLS)			/* add in new right column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (VesselsImage[(r+r2)*COLS+c+radius])
			{
		    total+=RawImage[((r+r2)*COLS+c+radius)*3+1];
		    total_pixels++;
			}
		  }
		}
	  if (r-radius >= 0)			/* add in new top row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (VesselsImage[(r-radius)*COLS+c+c2])
			{
			total+=RawImage[((r-radius)*COLS+c+c2)*3+1];
			total_pixels++;
			}
		  }
		}
	  if (r+radius < ROWS)			/* add in new bottom row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (VesselsImage[(r+radius)*COLS+c+c2])
			{
			total+=RawImage[((r+radius)*COLS+c+c2)*3+1];
			total_pixels++;
			}
		  }
		}
	  }		  /* end while() loop to expand window */

	RadiusImage[r*COLS+c]=50; /* radius*3; */
	if (total/total_pixels > 255)
      LocalImage[r*COLS+c]=255;
    else
      LocalImage[r*COLS+c]=total/total_pixels;
	}
  }

for (r=0; r<ROWS*COLS; r++)
  if (RadiusImage[r] > 255)
	AvgImage[r]=255;
  else
	AvgImage[r]=(unsigned char)RadiusImage[r];
/* WriteImage("C:\\Images\\STARE\\radius.ppm",AvgImage,ROWS,COLS,1,1); */
for (r=0; r<ROWS*COLS; r++)
  AvgImage[r]=0;

		/* smooth the local blood vessel intensity image */
for (r=0; r<ROWS; r++)
  {
  sprintf(text,"Smoothing...  %d%% complete",(int)(100.0*(double)r/(double)ROWS));
  hDC=GetDC(MainWnd);
  TextOut(hDC,0,0,text,strlen(text));
  ReleaseDC(MainWnd,hDC);
  radius=0;	  /* must start over at each row */
  if (MaskImage[r*COLS+0])	/* create starting window (1x1 in size) */
	{
	total=LocalImage[r*COLS+0];
	total_pixels=1;
	}
  else
	total=total_pixels=0;
  for (c=0; c<COLS; c++)
	{
	if (c > 0)	  /* translate window by one column */
	  {
	  if (c-(radius+1) >= 0)  /* subtract old leftmost column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c-(radius+1)])
			{
			total-=LocalImage[(r+r2)*COLS+c-(radius+1)];
			total_pixels--;
			}
		  }
		}
	  if (c+radius < COLS)	  /* add in new rightmost column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c+radius])
			{
			total+=LocalImage[(r+r2)*COLS+c+radius];
			total_pixels++;
			}
		  }
		}
	  }

	if (MaskImage[r*COLS+c] == 0)
	  continue;				  /* no local average needed for background pixels */

	while (radius > RadiusImage[r*COLS+c]) /* contract window if it contains too many pixels */
	  {
	  if (c-radius >= 0)			/* subtract out left column */
		{
	    for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c-radius])
			{
		    total-=LocalImage[(r+r2)*COLS+c-radius];
		    total_pixels--;
			}
		  }
		}
	  if (c+radius < COLS)			/* subtract out right column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c+radius])
			{
		    total-=LocalImage[(r+r2)*COLS+c+radius];
		    total_pixels--;
			}
		  }
		}
	  if (r-radius >= 0)			/* subtract out top row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r-radius)*COLS+c+c2])
			{
			total-=LocalImage[(r-radius)*COLS+c+c2];
			total_pixels--;
			}
		  }
		}
	  if (r+radius < ROWS)			/* subtract out bottom row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r+radius)*COLS+c+c2])
			{
			total-=LocalImage[(r+radius)*COLS+c+c2];
			total_pixels--;
			}
		  }
		}
	  radius--;
	  }		  /* end while() loop to contract window */

	while (radius < RadiusImage[r*COLS+c]) /* expand window if it contains too few pixels */
	  {
	  radius++;
	  if (c-radius >= 0)			/* add in new left column */
		{
	    for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c-radius])
			{
		    total+=LocalImage[(r+r2)*COLS+c-radius];
		    total_pixels++;
			}
		  }
		}
	  if (c+radius < COLS)			/* add in new right column */
		{
		for (r2=-radius; r2<=radius; r2++)
		  {
		  if (r+r2 < 0  ||  r+r2 >= ROWS)
			continue;
		  if (MaskImage[(r+r2)*COLS+c+radius])
			{
		    total+=LocalImage[(r+r2)*COLS+c+radius];
		    total_pixels++;
			}
		  }
		}
	  if (r-radius >= 0)			/* add in new top row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r-radius)*COLS+c+c2])
			{
			total+=LocalImage[(r-radius)*COLS+c+c2];
			total_pixels++;
			}
		  }
		}
	  if (r+radius < ROWS)			/* add in new bottom row */
		{
		for (c2=-radius+1; c2<=radius-1; c2++)
		  {
		  if (c+c2 < 0  ||  c+c2 >= COLS)
			continue;
		  if (MaskImage[(r+radius)*COLS+c+c2])
			{
			total+=LocalImage[(r+radius)*COLS+c+c2];
			total_pixels++;
			}
		  }
		}
	  }		  /* end while() loop to expand window */

	if (total/total_pixels > 255)
      AvgImage[r*COLS+c]=255;
    else
      AvgImage[r*COLS+c]=total/total_pixels;
	}
  }
for (r=0; r<ROWS*COLS; r++)
  LocalImage[r]=AvgImage[r];


/* WriteImage("C:\\Images\\STARE\\avg.ppm",LocalImage,ROWS,COLS,1,1); */
if (RETURN_LOCAL_AVERAGE_IMAGE)	/* for seeing local average image */
  {
  for (r=0; r<ROWS*COLS; r++)
	AreaImage[r]=LocalImage[r];
  }
else  /* return equalized image */
  {
  for (r=0; r<ROWS; r++)
	{
    for (c=0; c<COLS; c++)
	  {
	  if (!MaskImage[r*COLS+c])
	    continue;
	  if ((int)RawImage[(r*COLS+c)*3+1]+global_average-(int)LocalImage[r*COLS+c] > 255)
	    AreaImage[r*COLS+c]=255;
	  else if ((int)RawImage[(r*COLS+c)*3+1]+global_average-(int)LocalImage[r*COLS+c] < 0)
	    AreaImage[r*COLS+c]=0;
	  else
	    AreaImage[r*COLS+c]=RawImage[(r*COLS+c)*3+1]+global_average-LocalImage[r*COLS+c];
	  }
	}
  }

if (SAVE_RGB_EQUALIZED_IMAGE) /* for seeing the full-color image equalized */
  {
  RGBEqualizedImage=(unsigned char *)calloc(ROWS*COLS*3,1);
  for (r=0; r<ROWS; r++)
	{
    for (c=0; c<COLS; c++)
	  {
	  if (!MaskImage[r*COLS+c])
	    continue;
	  for (i=0; i<3; i++)
		{
	    if ((int)RawImage[(r*COLS+c)*3+i]+global_average-LocalImage[r*COLS+c] > 255)
	      RGBEqualizedImage[(r*COLS+c)*3+i]=255;
	    else if ((int)RawImage[(r*COLS+c)*3+i]+global_average-LocalImage[r*COLS+c] < 0)
	      RGBEqualizedImage[(r*COLS+c)*3+i]=0;
	    else
	      RGBEqualizedImage[(r*COLS+c)*3+i]=RawImage[(r*COLS+c)*3+i]+global_average-LocalImage[r*COLS+c];
		}
	  }
	}
  WriteImage("C:\\Images\\STARE\\EqualRGB.ppm",RGBEqualizedImage,ROWS,COLS,3,1);
  free(RGBEqualizedImage);
  }


free(MaskImage);
free(LocalImage);
free(AvgImage);
free(RadiusImage);
}


